@@ -5,6 +5,7 @@ import (
 	"fmt"
 	"os"
 	"reflect"
+	"strings"
 	"time"
 
 	corev1 "k8s.io/api/core/v1"
@@ -21,9 +22,10 @@ import (
 	restclient "k8s.io/client-go/rest"
 	"k8s.io/client-go/tools/cache"
 	"k8s.io/client-go/util/workqueue"
-	"k8s.io/klog"
+	"k8s.io/klog/v2"
 
 	configapiv1 "github.com/openshift/api/config/v1"
+	operatorv1 "github.com/openshift/api/operator/v1"
 	configset "github.com/openshift/client-go/config/clientset/versioned"
 	configsetv1 "github.com/openshift/client-go/config/clientset/versioned/typed/config/v1"
 	configinformers "github.com/openshift/client-go/config/informers/externalversions"
@@ -34,9 +36,11 @@ import (
 	tunedset "github.com/openshift/cluster-node-tuning-operator/pkg/generated/clientset/versioned"
 	tunedinformers "github.com/openshift/cluster-node-tuning-operator/pkg/generated/informers/externalversions"
 	ntomf "github.com/openshift/cluster-node-tuning-operator/pkg/manifests"
+	"github.com/openshift/cluster-node-tuning-operator/pkg/tuned"
 	"github.com/openshift/cluster-node-tuning-operator/pkg/util"
 	"github.com/openshift/cluster-node-tuning-operator/version"
 
+	ign3types "github.com/coreos/ignition/v2/config/v3_1/types"
 	mcfgclientset "github.com/openshift/machine-config-operator/pkg/generated/clientset/versioned"
 	mcfginformers "github.com/openshift/machine-config-operator/pkg/generated/informers/externalversions"
 )
@@ -66,9 +70,7 @@ type Controller struct {
 	listers *ntoclient.Listers
 	clients *ntoclient.Clients
 
-	pod struct {
-		informerFactory kubeinformers.SharedInformerFactory
-		informer        corev1informers.PodInformer
+	pod, node struct {
 		informerEnabled bool
 		stopCh          chan struct{}
 	}
@@ -148,11 +150,50 @@ func (c *Controller) eventProcessor() {
 
 func (c *Controller) sync(key wqKey) error {
 	var (
-		cr  *tunedv1.Tuned
-		err error
+		cr           *tunedv1.Tuned
+		err, lastErr error
 	)
 	klog.V(2).Infof("sync(): Kind %s: %s/%s", key.kind, key.namespace, key.name)
 
+	if key.kind == wqKindTuned && key.name == tunedv1.TunedDefaultResourceName {
+		// default Tuned changed or a bootstrap event received
+		cr, err = c.syncTunedDefault()
+		if err != nil {
+			return fmt.Errorf("failed to sync default Tuned CR: %v", err)
+		}
+	} else {
+		cr, err = c.listers.TunedResources.Get(tunedv1.TunedDefaultResourceName)
+		if err != nil {
+			if errors.IsNotFound(err) {
+				cr, err = c.syncTunedDefault()
+				if err != nil {
+					return fmt.Errorf("failed to sync default Tuned CR: %v", err)
+				}
+			} else {
+				return fmt.Errorf("failed to get Tuned %s: %v", tunedv1.TunedDefaultResourceName, err)
+			}
+		}
+	}
+	// We have the default Tuned custom resource (cr)
+
+	switch cr.Spec.ManagementState {
+	case operatorv1.Force:
+		// Use the same logic as Managed.
+	case operatorv1.Managed, "":
+		// Managed means that the operator is actively managing its resources and trying to keep the component active.
+	case operatorv1.Removed:
+		// Removed means that the operator is actively managing its resources and trying to remove all traces of the component.
+		lastErr = c.removeResources()
+		goto out
+	case operatorv1.Unmanaged:
+		// Unmanaged means that the operator will not take any action related to the component.
+		goto out
+	default:
+		// This should never happen due to openAPIV3Schema checks.
+		klog.Warningf("unknown custom resource ManagementState: %s", cr.Spec.ManagementState)
+	}
+	// Operator is in Force or Managed state.
+
 	switch {
 	case key.kind == wqKindPod:
 		klog.V(2).Infof("sync(): Pod %s/%s", key.namespace, key.name)
@@ -219,31 +260,6 @@ func (c *Controller) sync(key wqKey) error {
 
 		return nil
 
-	default:
-	}
-
-	if key.kind == wqKindTuned && key.name == tunedv1.TunedDefaultResourceName {
-		// default Tuned changed or a bootstrap event received
-		cr, err = c.syncTunedDefault()
-		if err != nil {
-			return fmt.Errorf("failed to sync default Tuned CR: %v", err)
-		}
-	} else {
-		cr, err = c.listers.TunedResources.Get(tunedv1.TunedDefaultResourceName)
-		if err != nil {
-			if errors.IsNotFound(err) {
-				cr, err = c.syncTunedDefault()
-				if err != nil {
-					return fmt.Errorf("failed to sync default Tuned CR: %v", err)
-				}
-			} else {
-				return fmt.Errorf("failed to get Tuned %s: %v", tunedv1.TunedDefaultResourceName, err)
-			}
-		}
-	}
-	// We have the default Tuned custom resource (cr)
-
-	switch {
 	case key.kind == wqKindDaemonSet, key.kind == wqKindClusterOperator:
 		klog.V(2).Infof("sync(): DaemonSet/OperatorStatus")
 
@@ -251,7 +267,7 @@ func (c *Controller) sync(key wqKey) error {
 		if err != nil {
 			return fmt.Errorf("failed to sync DaemonSet: %v", err)
 		}
-		err = c.syncOperatorStatus()
+		err = c.syncOperatorStatus(cr)
 		if err != nil {
 			return fmt.Errorf("failed to sync OperatorStatus: %v", err)
 		}
@@ -276,42 +292,66 @@ func (c *Controller) sync(key wqKey) error {
 	default:
 	}
 
-	// Tuned CR changed
+	// Tuned CR changed and the operator components need to be managed.
 	klog.V(2).Infof("sync(): Tuned %s", tunedv1.TunedRenderedResourceName)
 	err = c.syncTunedRendered(cr)
 	if err != nil {
 		return fmt.Errorf("failed to sync Tuned %s: %v", tunedv1.TunedRenderedResourceName, err)
 	}
 
+	if key.name != tunedv1.TunedDefaultResourceName {
+		crTuned, err := c.listers.TunedResources.Get(key.name)
+		if err != nil {
+			if !errors.IsNotFound(err) {
+				return err
+			}
+		} else if crTuned.Spec.ManagementState != "" {
+			klog.Warningf("setting ManagementState is supported only in Tuned/%s; ignoring ManagementState in Tuned/%s", tunedv1.TunedDefaultResourceName, key.name)
+		}
+	}
+
 	klog.V(2).Infof("sync(): DaemonSet")
 	err = c.syncDaemonSet(cr)
 	if err != nil {
 		return fmt.Errorf("failed to sync DaemonSet: %v", err)
 	}
 
-	if key.kind == wqKindTuned {
-		// Tuned CR changed, this can affect all profiles, list them and trigger profile updates
-		klog.V(2).Infof("sync(): Tuned %s", key.name)
+	// Tuned CR changed, this can affect all profiles, list them and trigger profile updates
+	klog.V(2).Infof("sync(): Tuned %s", key.name)
 
-		err = c.enqueueProfileUpdates()
-		if err != nil {
-			return err
-		}
+	err = c.enqueueProfileUpdates()
+	if err != nil {
+		return err
+	}
 
-		if key.name == tunedv1.TunedRenderedResourceName {
-			// Do not start unused MachineConfig pruning unnecessarily for the rendered resource
-			return nil
-		}
+	if key.name == tunedv1.TunedRenderedResourceName {
+		// Do not start unused MachineConfig pruning unnecessarily for the rendered resource
+		return nil
+	}
 
-		// Tuned CR change can also mean some MachineConfigs the operator created are no longer needed;
-		// removal of these will also rollback host settings such as kernel boot parameters.
-		err = c.pruneMachineConfigs()
-		if err != nil {
-			return err
-		}
+	// Tuned CR change can also mean some MachineConfigs the operator created are no longer needed;
+	// removal of these will also rollback host settings such as kernel boot parameters.
+	err = c.pruneMachineConfigs()
+	if err != nil {
+		return err
 	}
 
 	return nil
+
+out:
+	err = c.enableNodeInformer(false)
+	if err != nil {
+		lastErr = fmt.Errorf("failed to disable Node informer: %v", err)
+	}
+	err = c.enablePodInformer(false)
+	if err != nil {
+		lastErr = fmt.Errorf("failed to disable Pod informer: %v", err)
+	}
+	err = c.syncOperatorStatus(cr)
+	if err != nil {
+		lastErr = fmt.Errorf("failed to synchronize Operator status: %v", err)
+	}
+	return lastErr
 }
 
 // enqueueProfileUpdates enqueues profile calculations/updates for all Tuned Profiles in the cluster.
@@ -346,12 +386,14 @@ func (c *Controller) syncTunedDefault() (*tunedv1.Tuned, error) {
 	}
 
 	// Tuned resource found, check whether we need to update it
-	if reflect.DeepEqual(crMf.Spec, cr.Spec) {
+	if reflect.DeepEqual(crMf.Spec.Profile, cr.Spec.Profile) &&
+		reflect.DeepEqual(crMf.Spec.Recommend, cr.Spec.Recommend) {
 		klog.V(2).Infof("syncTunedDefault(): Tuned %s doesn't need updating", crMf.Name)
 		return cr, nil
 	}
 	cr = cr.DeepCopy() // never update the objects from cache
-	cr.Spec = crMf.Spec
+	cr.Spec.Profile = crMf.Spec.Profile
+	cr.Spec.Recommend = crMf.Spec.Recommend
 
 	klog.V(2).Infof("syncTunedDefault(): updating Tuned %s", crMf.Name)
 	cr, err = c.clients.Tuned.TunedV1().Tuneds(ntoconfig.OperatorNamespace()).Update(context.TODO(), cr, metav1.UpdateOptions{})
@@ -371,6 +413,9 @@ func (c *Controller) syncTunedRendered(tuned *tunedv1.Tuned) error {
 	crMf.ObjectMeta.OwnerReferences = getDefaultTunedRefs(tuned)
 	crMf.Name = tunedv1.TunedRenderedResourceName
 
+	nodeLabelsUsed := c.pc.tunedsUseNodeLabels(tunedList)
+	c.enableNodeInformer(nodeLabelsUsed)
+
 	// Enable/Disable Pod events based on tuned CRs using this functionality.
 	// It is strongly advised not to use the Pod-label functionality in large-scale clusters.
 	podLabelsUsed := c.pc.tunedsUsePodLabels(tunedList)
@@ -497,7 +542,7 @@ func (c *Controller) syncProfile(tuned *tunedv1.Tuned, nodeName string) error {
 		// The tuned profile "tunedProfileName" for nodeName matched with MachineConfig
 		// labels set for additional machine configuration.  Sync the operator-created
 		// MachineConfig for MachineConfigPools 'pools'.
-		err := c.syncMachineConfig(getMachineConfigNameForPools(pools), mcLabels, profile.Status.Bootcmdline)
+		err := c.syncMachineConfig(getMachineConfigNameForPools(pools), mcLabels, profile.Status.Bootcmdline, profile.Status.Stalld)
 		if err != nil {
 			return fmt.Errorf("failed to update Profile %s: %v", profile.Name, err)
 		}
@@ -519,10 +564,35 @@ func (c *Controller) syncProfile(tuned *tunedv1.Tuned, nodeName string) error {
 	return nil
 }
 
-func (c *Controller) syncMachineConfig(name string, labels map[string]string, bootcmdline string) error {
-	var kernelArguments []string
+func (c *Controller) syncMachineConfig(name string, labels map[string]string, bootcmdline string, stalld bool) error {
+	var (
+		kernelArguments []string
+		ignFiles        []ign3types.File
+		ignUnits        []ign3types.Unit
+	)
+	logline := func(bIgn, bCmdline bool, bootcmdline string) string {
+		var (
+			sb strings.Builder
+		)
 
+		if bIgn {
+			sb.WriteString(" ignition")
+			if bCmdline {
+				sb.WriteString(" and")
+			}
+		}
+
+		if bCmdline {
+			sb.WriteString(" kernel parameters: [")
+			sb.WriteString(bootcmdline)
+			sb.WriteString("]")
+		}
+
+		return sb.String()
+	}
 	kernelArguments = util.SplitKernelArguments(bootcmdline)
+	ignFiles = tuned.ProvideIgnitionFiles(stalld)
+	ignUnits = tuned.ProvideSystemdUnits(stalld)
 
 	annotations := map[string]string{GeneratedByControllerVersionAnnotationKey: version.Version}
 
@@ -530,37 +600,48 @@ func (c *Controller) syncMachineConfig(name string, labels map[string]string, bo
 	if err != nil {
 		if errors.IsNotFound(err) {
 			klog.V(2).Infof("syncMachineConfig(): MachineConfig %s not found, creating one", name)
-			mc = newMachineConfig(name, annotations, labels, kernelArguments)
-			if len(bootcmdline) == 0 {
-				// Creating a new MachineConfig with empty kernelArguments only causes unnecessary node
+			haveIgnition := len(ignFiles) != 0 || len(ignUnits) != 0
+			if len(bootcmdline) == 0 && !haveIgnition {
+				// Creating a new MachineConfig with empty kernelArguments/Ignition only causes unnecessary node
 				// reboots.
-				klog.V(2).Infof("not creating a MachineConfig with empty kernelArguments")
+				klog.V(2).Infof("not creating a MachineConfig with empty kernelArguments/Ignition")
 				return nil
 			}
+			mc = newMachineConfig(name, annotations, labels, kernelArguments, ignFiles, ignUnits)
 			_, err = c.clients.MC.MachineconfigurationV1().MachineConfigs().Create(context.TODO(), mc, metav1.CreateOptions{})
 			if err != nil {
 				return fmt.Errorf("failed to create MachineConfig %s: %v", mc.ObjectMeta.Name, err)
 			}
-			klog.Infof("created MachineConfig %s with kernelArguments: %v", mc.ObjectMeta.Name, kernelArguments)
+			klog.Infof("created MachineConfig %s with%s", mc.ObjectMeta.Name, logline(haveIgnition, len(bootcmdline) != 0, bootcmdline))
 			return nil
 		}
 		return err
 	}
 
-	if util.StringSlicesAsSetsEqual(mc.Spec.KernelArguments, kernelArguments) {
+	mcNew := newMachineConfig(name, annotations, labels, kernelArguments, ignFiles, ignUnits)
+
+	ignEq, err := ignEqual(mc, mcNew)
+	if err != nil {
+		return fmt.Errorf("failed to sync MachineConfig %s: %v", mc.ObjectMeta.Name, err)
+	}
+	kernelArgsEq := util.StringSlicesEqual(mc.Spec.KernelArguments, kernelArguments)
+	if kernelArgsEq && ignEq {
 		// No update needed
 		klog.V(2).Infof("syncMachineConfig(): MachineConfig %s doesn't need updating", mc.ObjectMeta.Name)
 		return nil
 	}
-
+	mc = mc.DeepCopy() // never update the objects from cache
 	mc.Spec.KernelArguments = kernelArguments
+	mc.Spec.Config = mcNew.Spec.Config
 
-	klog.V(2).Infof("syncMachineConfig(): updating MachineConfig %s with kernelArguments: %s", mc.ObjectMeta.Name, bootcmdline)
+	l := logline(!ignEq, !kernelArgsEq, bootcmdline)
+	klog.V(2).Infof("syncMachineConfig(): updating MachineConfig %s with%s", mc.ObjectMeta.Name, l)
 	_, err = c.clients.MC.MachineconfigurationV1().MachineConfigs().Update(context.TODO(), mc, metav1.UpdateOptions{})
 	if err != nil {
 		return fmt.Errorf("failed to update MachineConfig %s: %v", mc.ObjectMeta.Name, err)
 	}
-	klog.Infof("updated MachineConfig %s with kernelArguments: %v", mc.ObjectMeta.Name, kernelArguments)
+
+	klog.Infof("updated MachineConfig %s with%s", mc.ObjectMeta.Name, l)
 
 	return nil
 }
@@ -701,6 +782,35 @@ func (c *Controller) informerEventHandler(workqueueKey wqKey) cache.ResourceEven
 	}
 }
 
+// enableNodeInformer enables/disables event handling for Nodes.
+func (c *Controller) enableNodeInformer(enable bool) error {
+	if (enable && c.node.informerEnabled) || (!enable && !c.node.informerEnabled) {
+		return nil
+	}
+
+	if enable {
+		var (
+			informerFactory kubeinformers.SharedInformerFactory
+			informer        corev1informers.NodeInformer
+		)
+		c.node.stopCh = make(chan struct{})
+		informerFactory = kubeinformers.NewSharedInformerFactoryWithOptions(c.clients.Kube, ntoconfig.ResyncPeriod(), kubeinformers.WithNamespace(corev1.NamespaceAll))
+
+		informer = informerFactory.Core().V1().Nodes()
+		c.listers.Nodes = informer.Lister()
+		informer.Informer().AddEventHandler(c.informerEventHandler(wqKey{kind: wqKindNode}))
+
+		informerFactory.Start(c.node.stopCh)
+	} else {
+		defer close(c.node.stopCh)
+		c.node.stopCh <- struct{}{}
+		c.pc.nodeLabelsDelete()
+	}
+
+	c.node.informerEnabled = enable
+	return nil
+}
+
 // enablePodInformer enables/disables event handling for Pods.
 func (c *Controller) enablePodInformer(enable bool) error {
 	if (enable && c.pod.informerEnabled) || (!enable && !c.pod.informerEnabled) {
@@ -708,14 +818,18 @@ func (c *Controller) enablePodInformer(enable bool) error {
 	}
 
 	if enable {
+		var (
+			informerFactory kubeinformers.SharedInformerFactory
+			informer        corev1informers.PodInformer
+		)
 		c.pod.stopCh = make(chan struct{})
-		c.pod.informerFactory = kubeinformers.NewSharedInformerFactoryWithOptions(c.clients.Kube, ntoconfig.ResyncPeriod(), kubeinformers.WithNamespace(corev1.NamespaceAll))
+		informerFactory = kubeinformers.NewSharedInformerFactoryWithOptions(c.clients.Kube, ntoconfig.ResyncPeriod(), kubeinformers.WithNamespace(corev1.NamespaceAll))
 
-		c.pod.informer = c.pod.informerFactory.Core().V1().Pods()
-		c.listers.Pods = c.pod.informer.Lister()
-		c.pod.informer.Informer().AddEventHandler(c.informerEventHandler(wqKey{kind: wqKindPod}))
+		informer = informerFactory.Core().V1().Pods()
+		c.listers.Pods = informer.Lister()
+		informer.Informer().AddEventHandler(c.informerEventHandler(wqKey{kind: wqKindPod}))
 
-		c.pod.informerFactory.Start(c.pod.stopCh)
+		informerFactory.Start(c.pod.stopCh)
 	} else {
 		defer close(c.pod.stopCh)
 		c.pod.stopCh <- struct{}{}
@@ -726,6 +840,60 @@ func (c *Controller) enablePodInformer(enable bool) error {
 	return nil
 }
 
+func (c *Controller) removeResources() error {
+	var lastErr error
+	dsMf := ntomf.TunedDaemonSet()
+	ctx := context.TODO()
+
+	_, err := c.listers.DaemonSets.Get(dsMf.Name)
+	if err != nil {
+		if !errors.IsNotFound(err) {
+			lastErr = fmt.Errorf("failed to get DaemonSet %s: %v", dsMf.Name, err)
+		}
+	} else {
+		err = c.clients.Apps.DaemonSets(ntoconfig.OperatorNamespace()).Delete(ctx, dsMf.Name, metav1.DeleteOptions{})
+		if err != nil {
+			lastErr = fmt.Errorf("failed to delete DaemonSet %s: %v", dsMf.Name, err)
+		} else {
+			klog.Infof("deleted DaemonSet %s", dsMf.Name)
+		}
+	}
+
+	_, err = c.listers.TunedResources.Get(tunedv1.TunedRenderedResourceName)
+	if err != nil {
+		if !errors.IsNotFound(err) {
+			lastErr = fmt.Errorf("failed to get Tuned %s: %v", tunedv1.TunedRenderedResourceName, err)
+		}
+	} else {
+		err = c.clients.Tuned.TunedV1().Tuneds(ntoconfig.OperatorNamespace()).Delete(ctx, tunedv1.TunedRenderedResourceName, metav1.DeleteOptions{})
+		if err != nil {
+			lastErr = fmt.Errorf("failed to delete Tuned %s: %v", tunedv1.TunedRenderedResourceName, err)
+		} else {
+			klog.Infof("deleted Tuned %s", tunedv1.TunedRenderedResourceName)
+		}
+	}
+
+	profileList, err := c.listers.TunedProfiles.List(labels.Everything())
+	if err != nil {
+		lastErr = fmt.Errorf("failed to list Tuned profiles: %v", err)
+	}
+	for _, profile := range profileList {
+		err = c.clients.Tuned.TunedV1().Profiles(ntoconfig.OperatorNamespace()).Delete(ctx, profile.Name, metav1.DeleteOptions{})
+		if err != nil && !errors.IsNotFound(err) {
+			lastErr = fmt.Errorf("failed to delete Profile %s: %v", profile.Name, err)
+		} else {
+			klog.Infof("deleted Profile %s", profile.Name)
+		}
+	}
+
+	err = c.pruneMachineConfigs()
+	if err != nil {
+		lastErr = fmt.Errorf("failed to prune operator-created MachineConfigs: %v", err)
+	}
+
+	return lastErr
+}
+
 // Run will set up the event handlers for types we are interested in, as well
 // as syncing informer caches and starting workers. It will block until stopCh
 // is closed, at which point it will shutdown the workqueue and wait for
@@ -793,14 +961,9 @@ func (c *Controller) Run(stopCh <-chan struct{}) error {
 
 	configInformerFactory := configinformers.NewSharedInformerFactory(configClient, ntoconfig.ResyncPeriod())
 	kubeNTOInformerFactory := kubeinformers.NewSharedInformerFactoryWithOptions(c.clients.Kube, ntoconfig.ResyncPeriod(), kubeinformers.WithNamespace(ntoconfig.OperatorNamespace()))
-	kubeInformerFactory := kubeinformers.NewSharedInformerFactoryWithOptions(c.clients.Kube, ntoconfig.ResyncPeriod(), kubeinformers.WithNamespace(corev1.NamespaceAll))
 	tunedInformerFactory := tunedinformers.NewSharedInformerFactoryWithOptions(c.clients.Tuned, ntoconfig.ResyncPeriod(), tunedinformers.WithNamespace(ntoconfig.OperatorNamespace()))
 	mcfgInformerFactory := mcfginformers.NewSharedInformerFactory(c.clients.MC, ntoconfig.ResyncPeriod())
 
-	nodeInformer := kubeInformerFactory.Core().V1().Nodes()
-	c.listers.Nodes = nodeInformer.Lister()
-	nodeInformer.Informer().AddEventHandler(c.informerEventHandler(wqKey{kind: wqKindNode}))
-
 	coInformer := configInformerFactory.Config().V1().ClusterOperators()
 	c.listers.ClusterOperators = coInformer.Lister()
 	coInformer.Informer().AddEventHandler(c.informerEventHandler(wqKey{kind: wqKindClusterOperator}))
@@ -826,14 +989,12 @@ func (c *Controller) Run(stopCh <-chan struct{}) error {
 
 	configInformerFactory.Start(stopCh)  // ClusterOperator
 	kubeNTOInformerFactory.Start(stopCh) // DaemonSet
-	kubeInformerFactory.Start(stopCh)    // Node
 	tunedInformerFactory.Start(stopCh)   // Tuned/Profile
 	mcfgInformerFactory.Start(stopCh)    // MachineConfig/MachineConfigPool
 
 	// Wait for the caches to be synced before starting worker(s)
 	klog.V(1).Info("waiting for informer caches to sync")
 	ok := cache.WaitForCacheSync(stopCh,
-		nodeInformer.Informer().HasSynced,
 		coInformer.Informer().HasSynced,
 		dsInformer.Informer().HasSynced,
 		trInformer.Informer().HasSynced,
@@ -850,6 +1011,7 @@ func (c *Controller) Run(stopCh <-chan struct{}) error {
 	klog.Info("started events processor/controller")
 
 	<-stopCh
+	c.enableNodeInformer(false)
 	c.enablePodInformer(false)
 	klog.Info("shutting down events processor/controller")
 
